#!/bin/python
# -*- coding: utf-8 -*-
"""
build_help.py

Author: GrimAndGreedy
License: MIT
"""

from listpick.ui.keys import picker_keys, notification_keys
import curses
import logging
from listpick.utils import keycodes

logger = logging.getLogger('picker_log')

def build_help_rows(keys_dict: dict, debug: bool = False) -> list[list[str]]:
    """ Build help rows based on the keys_dict. """

    logger.info(f"function: build_help_rows() (build_help.py)")

    ## Key names
    special_keys = {

            27: "Escape",
            353: "Shift+Tab",
            curses.KEY_END: "END",
            curses.KEY_HOME: "HOME",
            curses.KEY_PPAGE: "Page Up",
            curses.KEY_NPAGE: "Page Down",
            curses.KEY_UP: "ArrowUp",
            curses.KEY_DOWN: "ArrowDown",
            curses.KEY_RIGHT: "ArrowRight",
            curses.KEY_LEFT: "ArrowLeft",
            ord(' '): "Space",
            curses.KEY_ENTER: "RETURN",
            ord('\n'): "\n",
            curses.KEY_DC: "Delete",
            383: "Shift+Delete",
            ord("\t"): "Tab",
            curses.KEY_BACKSPACE: "Backspace",
            keycodes.META_BS: "Alt+Backspace",
    }

    # Ctrl + [a-z]
    for i in range(26):
        special_keys[i+1] = f"Ctrl+{chr(ord('a')+i)}"
        if i == 8:
            special_keys[i+1] = f"Tab/Ctrl+{chr(ord('a')+i)}"


    # F1-F12
    for i in range(12):
        special_keys[curses.KEY_F1+i] = f"F{i+1}"

    # Alt+[a-z]
    for i in range(26):
        special_keys[keycodes.META_a +i] = f"Alt+{chr(ord('a')+i)}"

    # Alt+[A-Z]
    for i in range(26):
        special_keys[keycodes.META_A +i] = f"Alt+{chr(ord('A')+i)}"

    # Alt+[0-9]
    for i in range(10):
        special_keys[keycodes.META_0] = f"Alt+{i}"


    ## Key descriptions
    help_descriptions = {
        "refresh":                          "Refresh the screen.",
        "help":                             "Open help.",
        "exit":                             "Exit picker instance.",
        "full_exit":                        "Immediate exit to terminal.",
        "move_column_left":                 "Move column left.",
        "move_column_right":                "Move column right.",
        "cursor_down":                      "Cursor down.",
        "cursor_up":                        "Cursor up.",
        "half_page_up":                     "Half page up.",
        "half_page_down":                   "Half page down.",
        "page_up":                          "Page up.",
        "page_down":                        "Page down.",
        "cursor_bottom":                    "Send cursor to bottom of list.",
        "cursor_top":                       "Send cursor to top of list.",
        "five_up":                          "Five up.",
        "five_down":                        "Five down.",
        "toggle_select":                    "Toggle selection.",
        "select_all":                       "Select all.",
        "select_none":                      "Select none.",
        "visual_selection_toggle":          "Toggle visual selection.",
        "visual_deselection_toggle":        "Toggle visual deselection.",
        "enter":                            "Accept selections.",
        "redraw_screen":                    "Redraw screen.",
        "cycle_sort_method":                "Cycle through sort methods.",
        "cycle_sort_method_reverse":        "Cycle through sort methods (reverse)",
        "cycle_sort_order":                 "Toggle sort order.",
        "delete":                           "Delete row.",
        "delete_column":                    "Delete column.",
        "decrease_lines_per_page":          "Decrease lines per page.",
        "increase_lines_per_page":          "Increase lines per page.",
        "increase_column_width":            "Increase column width.",
        "decrease_column_width":            "Decrease column width.",
        "filter_input":                     "Filter rows.",
        "search_input":                     "Search.",
        "settings_input":                   "Settings input.",
        "settings_options":                 "Settings options dialogue.",
        "continue_search_forward":          "Continue search forwards.",
        "continue_search_backward":         "Continue search backwards.",
        "cancel":                           "Cancel; escape.",
        "opts_input":                       "Options input.",
        "opts_select":                      "Options select dialogue.",
        "mode_next":                        "Cycle through modes forwards.",
        "mode_prev":                        "Cycle through modes backwards.",
        "pipe_input":                       "Pipe selected cells from selected rows.",
        "reset_opts":                       "Reset options.",
        "col_select":                       "Select column.",
        "col_select_next":                  "Select next column.",
        "col_select_prev":                  "Select previous column.",
        "col_hide":                         "Hide column.",
        "edit":                             "Edit cell.",
        "edit_picker":                      "Edit cell from options dialogue.",
        "edit_ipython":                     "Edit current data with ipython.",
        "copy":                             "Copy selections.",
        "paste":                            "Paste into picker.",
        "save":                             "Save selections.",
        "load":                             "Load from file.",
        "open":                             "Attempt to open file with selected cell value(s).",
        "toggle_footer":                    "Toggle footer.",
        "notification_toggle":              "Toggle empty notification.",
        "redo":                             "Redo.",
        "undo":                             "Undo.",
        "scroll_right":                     "Scroll right.",
        "scroll_left":                      "Scroll left.",
        "scroll_far_right":                 "Scroll to the end of the column set.",
        "scroll_far_left":                  "Scroll to the left home.",
        "add_column_before":                "Insert column before cursor.",
        "add_row_before":                   "Insert row before cursor.",
        "add_column_after":                 "Insert column after cursor.",
        "add_row_after":                    "Insert row after cursor.",
        "info":                             "Display info screen.",
        "file_next":                        "Go to the next open file.",
        "file_prev":                        "Go to the previous open file.",
        "sheet_next":                       "Go to the next sheet.",
        "sheet_prev":                       "Go to the previous sheet.",
        "toggle_right_pane":                "Toggle the right pane",
        "cycle_right_pane":                 "Cycle through right panes",
    }
    sections = {
        "Navigation:": [ "cursor_down", "cursor_up", "half_page_up", "half_page_down", "page_up", "page_down", "cursor_bottom", "cursor_top", "five_up", "five_down", "scroll_right", "scroll_left", "scroll_far_right", "scroll_far_left" ],
        "Selection:": [ "toggle_select", "select_all", "select_none", "visual_selection_toggle", "visual_deselection_toggle", "enter" ],
        "UI:": [ "toggle_footer", "redraw_screen", "decrease_lines_per_page", "increase_lines_per_page", "increase_column_width", "decrease_column_width", "notification_toggle", "toggle_right_pane", "cycle_right_pane"],
        "Sort:": [ "cycle_sort_method", "cycle_sort_method_reverse", "cycle_sort_order", ] ,
        "Data manipulation:": [ "delete", "delete_column", "edit", "edit_picker", "edit_ipython", "add_column_before", "add_column_after", "add_row_before", "add_row_after"],
        "Filter and sort:": [ "filter_input", "search_input", "continue_search_forward", "continue_search_backward", ] ,
        "Settings:": [ "settings_input", "settings_options" ],
        "Cancel:": [ "opts_input", "opts_select", "mode_next", "mode_prev", "pipe_input", "reset_opts", "col_select", "col_select_next", "col_select_prev", "col_hide" ],
        "Save, load, copy and paste:": [ "save", "load", "open", "copy", "paste" ],
        "Misc:": [ "redo", "undo", "refresh", "help", "exit", "full_exit", "move_column_left", "move_column_right" ],
    }

    ## Add any keys not in section keys to misc.
    for key, desc in help_descriptions.items():
        found = False
        for section in sections:
            if key in sections[section]:
                found = True
                break
        if not found:
            sections["Misc:"].append(key)

    items = []
    for section_name, section_operations in sections.items():
        section_rows = []
        
        for operation in section_operations:
            keys = []
            if operation in keys_dict:
                for key in keys_dict[operation]:
                    if key in special_keys:
                        keys.append(special_keys[key])
                    else:
                        try:
                            keys.append(chr(int(key)))
                        except Exception as e:
                            keys.append(f"keycode={key}")
                            if debug: print(f"Error chr({key}): {e}")
            else:
                if debug: print(f"Note that {operation} is not in the keys_dict")
                continue



            if operation in help_descriptions:
                description = help_descriptions[operation]
            else:
                if debug: print(f"Operation={operation} has no description.")
                description = "..."

            row = [f"    {str(keys)[1:-1]}", description]
            section_rows.append(row)
        if section_rows:
            items.append([f"  {section_name}", ""])
            items += section_rows
            items.append(["",""])

    if debug:
        for operation in keys_dict:
            if operation not in help_descriptions:
                print(f"Note that {operation} is not in the help_descriptions")

    return items

if __name__ == "__main__":
    items = build_help_rows(picker_keys, debug=True)
    # items = build_help_rows(notification_keys, debug=True)

    # from listpick.listpick_app import Picker, start_curses, close_curses
    # stdscr = start_curses()
    # x = Picker(
    #     stdscr,
    #     items=items
    # )
    # x.run()
    #
    # close_curses(stdscr)



#!/bin/python
# -*- coding: utf-8 -*-
"""
graphing.py

Author: GrimAndGreedy
License: MIT
"""

import sys, os
sys.path.append(os.path.expanduser(".."))
os.chdir(os.path.dirname(os.path.realpath(__file__)))
os.chdir("../../..")
from aria2tui.lib.aria2c_wrapper import *
from aria2tui.utils.aria2c_utils import *
from listpick import *
from listpick.listpick_app import *
import time
import re
from typing import Callable

def escape_ansi(line: str) -> str:
    """ Remove ansi characters from string. """
    ansi_escape =re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
    return ansi_escape.sub('', line)

def handle_plotille_not_found(stdscr: curses.window) -> None:
    """ Display ModuleNotFoundError. """
    h, w = stdscr.getmaxyx()
    s = "ModuleNotFoundError: No module named 'plotille'"
    stdscr.addstr(h//2, (w - len(s))//2, s)
    stdscr.refresh()
    stdscr.getch()


def graph_speeds(
        stdscr: curses.window,
        get_data_function: Callable,
        timeout:int=1000,
        title:str="",
        refresh_time: int = 2,
        xposf: Callable = lambda: 0,
        yposf: Callable = lambda: 0,
        graph_wh: Callable[None,Tuple[int, int]] = lambda: os.get_terminal_size(),

    ) -> None:
    """ Display a graph of the global stats in a curses window. """
    try:
        import plotille as plt
    except:
        handle_plotille_not_found(stdscr)
        return None

    initial_time = time.time()
    x, y, y2 = [], [], []
    while time.time()-initial_time < timeout:


        resp = get_data_function()
        x.append(time.time()-initial_time)
        down = int(resp['result']['downloadSpeed'])
        up = int(resp['result']['uploadSpeed'])

        y.append(down)
        y2.append(up)

        fig = plt.Figure()
        fig.plot(x, y)
        fig.plot(x, y2)

        fig.y_ticks_fkt = lambda y, _: bytes_to_human_readable(y)
        fig.x_ticks_fkt = lambda x, _: f"{int(x)}s"
        fig.set_y_limits(min_=0)
        fig.set_x_limits(min_=0)
        fig.text([x[-1]], [y[0]], ['Dn'])
        fig.text([x[0]], [y2[0]], ['Up'])

        width, height = graph_wh()
        fig.width = width - 7
        fig.height = height - 4
        globh, globw = stdscr.getmaxyx()
        xpos, ypos = xposf(), yposf()
        maxw, maxh = globw-xpos-1, globh-ypos

        stdscr.erase()

        # Draw title
        stdscr.addstr(ypos, xpos, f"{title:^{min(width,maxw)}}")

        # Draw graph
        lines = fig.show().split('\n')
        for i, line in enumerate(lines):
            if i > maxh-2: break
            line = escape_ansi(line)

            stdscr.addstr(i+1+ypos, xpos , line[:min(width,maxw)])

        # Show the extreme points of the width and also the last printable char
        show_control_chars = False
        if show_control_chars:
            stdscr.addstr(0,0, f"{maxw}, {maxh}")
            stdscr.addstr(ypos, xpos, "*", curses.A_REVERSE)
            stdscr.addstr(ypos+maxh-1, xpos, "*", curses.A_REVERSE)
            stdscr.addstr(ypos, xpos+maxw-1, "*", curses.A_REVERSE)
            stdscr.addstr(ypos+maxh-1, xpos+maxw-2, "*", curses.A_REVERSE)
            stdscr.addstr(ypos+ min(height, maxh-1), xpos, "+", curses.A_REVERSE)
            stdscr.addstr(ypos, xpos+min(maxw-1, width), "+", curses.A_REVERSE)
            stdscr.addstr(ypos+min(height, maxh-1), xpos+min(width, maxw-2), "+", curses.A_REVERSE)

        stdscr.refresh()
        key = stdscr.getch()
        if key in [3, ord('q')]: 
            return None

def graph_speeds_gid(
        stdscr: curses.window,
        get_data_function: Callable,
        timeout:int=1000,
        title:str="",
        refresh_time: int = 2,
        xposf: Callable = lambda: 0,
        yposf: Callable = lambda: 0,
        graph_wh: Callable[None,Tuple[int, int]] = lambda: os.get_terminal_size(),
        gid:str = ""

    ) -> None:
    """ Display a graph in a curses window for a certain gid. """
    try:
        import plotille as plt
    except:
        handle_plotille_not_found(stdscr)
        return None

    initial_time = time.time()
    x, y, y2 = [], [], []
    while time.time()-initial_time < timeout:

        resp = get_data_function(gid)
        x.append(time.time()-initial_time)
        down = int(resp['result']['downloadSpeed'])
        up = int(resp['result']['uploadSpeed'])

        y.append(down)
        y2.append(up)

        fig = plt.Figure()
        fig.plot(x, y)
        fig.plot(x, y2)

        fig.y_ticks_fkt = lambda y, _: bytes_to_human_readable(y)
        fig.x_ticks_fkt = lambda x, _: f"{int(x)}s"
        fig.set_y_limits(min_=0)
        fig.set_x_limits(min_=0)
        fig.text([x[-1]], [y[0]], ['Dn'])
        fig.text([x[0]], [y2[0]], ['Up'])

        width, height = graph_wh()
        fig.width = width - 7
        fig.height = height - 4
        globh, globw = stdscr.getmaxyx()
        xpos, ypos = xposf(), yposf()
        maxw, maxh = globw-xpos-1, globh-ypos

        stdscr.erase()

        # Draw title
        stdscr.addstr(ypos, xpos, f"{title:^{min(width,maxw)}}")

        # Draw graph
        lines = fig.show().split('\n')
        for i, line in enumerate(lines):
            if i > maxh-2: break
            line = escape_ansi(line)

            stdscr.addstr(i+1+ypos, xpos , line[:min(width,maxw)])

        # Show the extreme points of the width and also the last printable char
        show_control_chars = False
        if show_control_chars:
            stdscr.addstr(0,0, f"{maxw}, {maxh}")
            stdscr.addstr(ypos, xpos, "*", curses.A_REVERSE)
            stdscr.addstr(ypos+maxh-1, xpos, "*", curses.A_REVERSE)
            stdscr.addstr(ypos, xpos+maxw-1, "*", curses.A_REVERSE)
            stdscr.addstr(ypos+maxh-1, xpos+maxw-2, "*", curses.A_REVERSE)
            stdscr.addstr(ypos+ min(height, maxh-1), xpos, "+", curses.A_REVERSE)
            stdscr.addstr(ypos, xpos+min(maxw-1, width), "+", curses.A_REVERSE)
            stdscr.addstr(ypos+min(height, maxh-1), xpos+min(width, maxw-2), "+", curses.A_REVERSE)

        stdscr.refresh()
        key = stdscr.getch()
        stdscr.refresh()
        if key in [3, ord('q')]: 
            return None
        resp = get_data_function(gid)



if __name__ == "__main__":
    title = "Global Transfer Speeds"
    end_time = 180
    get_data_function = lambda: sendReq(getGlobalStat())
    gid = "60e0c88ed77a24d6"
    get_data_function = lambda: sendReq(tellStatus(gid))
    wait_time = 2

    stdscr = start_curses()
    size_func = lambda: (
        3*os.get_terminal_size()[0]//4,
        3*os.get_terminal_size()[1]//4,
    )
    xposf = lambda: os.get_terminal_size()[0]//8
    yposf = lambda: os.get_terminal_size()[1]//8
    graph_speeds(
        stdscr,
        get_data_function=get_data_function,
        timeout=end_time,
        refresh_time=wait_time,
        title=title,
        graph_wh= size_func,
        xposf=xposf,
        yposf=yposf,
            
    )
    close_curses(stdscr)



#!/bin/python
# -*- coding: utf-8 -*-
"""
Create a basic picker with a title, header, and items.

Author: GrimAndGreedy
License: MIT
"""

from listpick.listpick_app import Picker, close_curses, start_curses

l = [["1", "2"], ["3", "4"]]
header=["Pericles", "Is Dead"]

stdscr = start_curses()
x = Picker(
        stdscr = stdscr,
        items = l,
        title="Test Picker",
        header=header,
        # highlights=highlights,
    )
selected_indices, opts, picker_data = x.run()

close_curses(stdscr)

print(f"Selected: {selected_indices}")
print(f"Opts: {opts}")



#!/bin/python
# -*- coding: utf-8 -*-
"""
pane_utils.py
Utility functions for Picker panes.

Author: GrimAndGreedy
License: MIT
"""

import re

def escape_ansi(line: str) -> str:
    """ Remove ansi characters from string. """
    ansi_escape =re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
    return ansi_escape.sub('', line)


def get_graph_string(x_vals, y_vals, width=50, height=20, title=None, x_label=None, y_label=None):
    """ Generate a graph of x_vals, y_vals using plotille"""

    import plotille as plt
    # Create a figure and axis object using plotille
    fig = plt.Figure()
    
    # Plot the data on the figure
    fig.plot(x_vals, y_vals)
    
    # Set the dimensions of the graph
    fig.width = width-10
    fig.height = height-4
    fig.x_ticks_fkt = lambda x, _: f"{int(x)}s"
    fig.y_ticks_fkt = lambda y, _: f"{int(y)}"
    
    # Set the title and labels if provided
    if title:
        fig.title = title
    if x_label:
        fig.xlabel = x_label
    if y_label:
        fig.ylabel = y_label
    
    # Generate the ASCII art of the graph
    ascii_art = str(fig.show())
    
    return ascii_art


def get_file_attributes(filename):
    import mimetypes
    import time
    import os
    try:
        if not os.path.exists(filename) and len(filename) > 2 and filename[0] == "'" and filename[-1] == "'":
            filename = filename[1:-1] 

        abs_path = os.path.abspath(filename)
        
        if not os.path.exists(abs_path):
            return ["File not found"]

        # Get the size of the file in bytes and convert it to a human-readable format
        size = os.path.getsize(abs_path)
        size_str = f"{(size / (1024 * 1024 * 1024)):,.2f}GB" if size > 1024 ** 3 else \
                   f"{(size / (1024 * 1024)):,.2f}MB" if size > 1024 ** 2 else \
                   f"{(size / 1024):,.2f}KB" if size > 1024 else f"{size}B"
        
        # Get the file type
        mime_type, _ = mimetypes.guess_type(abs_path)
        if mime_type is None:
            mime_type = "application/octet-stream"
        
        # Get the last modified time in a human-readable format
        mod_time = os.path.getmtime(abs_path)
        mod_str = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(mod_time))
        
        attributes = [
            f"Size: {size_str}",
            f"Filetype: {mime_type}",
            f"Last Modified: {mod_str}"
        ]
        
        return attributes
    
    except Exception as e:
        print(f"An error occurred: {e}")
        return []




#!/bin/python
# -*- coding: utf-8 -*-
"""
search_and_filter_utils.py
Utilities for searching and filtering.

Author: GrimAndGreedy
License: MIT
"""

import re
import logging

logger = logging.getLogger('picker_log')

def apply_filter(row: list[str], filters: dict, case_sensitive: bool = False, add_highlights:bool = False, highlights: list=[]) -> bool:
    """ Checks if row matches the filter. """
    logger.info("function: apply_filter (search_and_filter_utils.py)")
    for col, filter_list in filters.items():
        for filter in filter_list:
            if case_sensitive or (filter != filter.lower()):
                pattern = re.compile(filter)
            else:
                pattern = re.compile(filter, re.IGNORECASE)
            if col == -1:  # Apply filter to all columns
                if not any(pattern.search(str(item)) for item in row):
                    return False
                # return not invert_filter
            elif col >= len(row) or col < 0:
                return False
            else:
                cell_value = str(row[col])
                if not pattern.search(cell_value):
                    return False
                # return invert_filter

    if add_highlights:
        for col, filter_list in filters.items():
            for filter in filter_list:
                hcol = "all" if col == -1 else col
                highlight = {
                    "match": filter,
                    "field": hcol,
                    "color": 10,
                    "type": "search",
                    "level": 1,
                }
                if highlight not in highlights:
                    highlights.append(highlight)
    
    return True


def tokenise(query:str) -> dict:
    """ Convert query into dict consisting of filters. '--1  """
    logger.info("function: tokenise (search_and_filter_utils.py)")
    filters = {}

    # tokens = re.split(r'(\s+--\d+|\s+--i)', query)
    # tokens = re.split(r'((\s+|^)--\w)', query)
    tokens = re.split(r'\s+', query)
    tokens = [token.strip() for token in tokens if token.strip()]  # Remove empty tokens
    i = 0
    while i < len(tokens):
        token = tokens[i]
        if token:
            if token.startswith("--"):
                flag = token
                if flag == '--v':
                    invert_filter = True 
                    i += 1
                elif flag == '--i':
                    case_sensitive = True
                    i += 1
                else:
                    if i+1 >= len(tokens):
                        break
                    col = int(flag[2:])
                    arg = tokens[i+1].strip()
                    try:
                        i+=2
                        re.compile(arg)
                        if col in filters: filters[col].append(arg)
                        else: filters[col] = [arg]
                    except:
                        pass
            else:
                try:
                    i += 1
                    re.compile(token)
                    if -1 in filters: filters[-1].append(token)
                    else: filters[-1] = [token]
                except:
                    pass
        else:
            i += 1
    return filters



"""
picker_log.py
Lines to be displayed on the help screen.

Author: GrimAndGreedy
License: MIT
"""

import logging


def setup_logger(name="picker_log", log_file="picker.log", log_enabled=True, level=logging.INFO):
    """ Set up a logger """
    logger = logging.getLogger(name)

    logger.handlers.clear()
    if log_enabled:
        # # prevent 
        # if not logger.handlers:
        logger.propagate = False
        file_handler = logging.FileHandler(log_file)        
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s',  '%m-%d-%Y %H:%M:%S')
        file_handler.setFormatter(formatter)

        logger.setLevel(level)
        logger.addHandler(file_handler)
        logger.disabled = False
    else:
        logger.disabled = True

    return logger

